#ifdef HAVE_CONFIG_H
#include <config.h>
#endif


// standard includes
#include <stdexcept>  // runtime_error 
#include <iostream>   // cout
#ifdef HAVE_SSTREAM
#include <sstream>
#else
#include <strstream>  // ostrstream, istrstream
#endif

// the general include for eo

#include <eo>
#include <ga.h>


#include "fitness.h"
#include "gui/gui.h"


using namespace std;


void main_function(int argc, char **argv)
{


   unsigned int SEED;                   // seed for random number generator
   unsigned int VEC_SIZE = 70;         // Number of bits in genotypes
   unsigned int POP_SIZE;               // Size of population
   unsigned int MAX_GEN;                // Maximum number of generation before STOP
   double CROSS_RATE;                   // Crossover rate
   double P_MUT_PER_BIT;                // probability of bit-flip mutation
   double MUT_RATE;                     // mutation rate


   // PARAMETRES

    eoParser parser(argc, argv);
    
    eoValueParam<uint32_t> seedParam(time(0), "seed", "Random number seed", 'S');
    parser.processParam( seedParam );
    SEED = seedParam.value();
    
    eoValueParam<unsigned int> popSizeParam(100, "popSize", "Population size",'P');
    parser.processParam( popSizeParam, "Evolution engine" );
    POP_SIZE = popSizeParam.value();

    eoValueParam<unsigned int> maxGenParam(500, "maxGen", "Maximum number of generations",'G');
    parser.processParam( maxGenParam, "Stopping criterion" );
    MAX_GEN = maxGenParam.value();

    eoValueParam<double> pCrossParam(0.6, "pCross", "Probability of Crossover", 'C'); 
    parser.processParam( pCrossParam, "Genetic Operators" );
    CROSS_RATE = pCrossParam.value();

    eoValueParam<double> pMutParam(0.4, "pMut", "Probability of Mutation", 'M');
    parser.processParam( pMutParam, "Genetic Operators" );
    MUT_RATE = pMutParam.value();

    eoValueParam<double> pMutPerBitParam(0.01, "pMutPerBit", "Probability of flipping 1 bit in bit-flip mutation", 'b');
    parser.processParam( pMutPerBitParam, "Genetic Operators" );
    P_MUT_PER_BIT = pMutPerBitParam.value();

    eoValueParam<string> saveStatsParam("prisoner.log", "LogFile","A file to log statistics",'L');
    parser.processParam( saveStatsParam, "Log Statistics File" );
    string saveStatsName = saveStatsParam.value();

    string str_status = parser.ProgramName() + ".status"; 
    eoValueParam<string> statusParam(str_status.c_str(), "status","Status file",'S');
    parser.processParam( statusParam, "Persistence" );

    
    if (parser.userNeedsHelp())
      {
        parser.printHelp(cout);
        exit(1);
      }
    if (statusParam.value() != "")
      {
	ofstream os(statusParam.value().c_str());
	os << parser;
      }


    //////////////////////////
    //  Random seed
    //////////////////////////
    rng.reseed(SEED);

    /////////////////////////////
    // Fitness function
    ////////////////////////////
    PrisonerDilemaFitness eval;


    myGui<Indi> gui(CROSS_RATE,MUT_RATE,
		    eval,
		    50,330);
    
    gui.do_select(); 

    ////////////////////////////////
    // Initilisation of population
    ////////////////////////////////
    eoPop<Indi> pop;
    for (unsigned int igeno=0; igeno<POP_SIZE; igeno++)
      {
	Indi v;           // void individual, to be filled
	for (unsigned ivar=0; ivar<VEC_SIZE; ivar++)
	  {
	    bool r = rng.flip(); // new value, random in {0,1}
	    v.push_back(r);      // append that random value to v
	  }
	//v.fitness( eval.compute_fitness(v) );                 // evaluate it
	pop.push_back(v);        // and put it in the population
      }
    eval(pop,pop);

    // OUTPUT
    pop.sort();
    cout << "Initial Population" << endl;
    cout << pop;

    /////////////////////////////////////
    // selection and replacement
    ////////////////////////////////////
    eoProportionalSelect<Indi> selectOne; 
    eoSelectPerc<Indi> select(selectOne); // by default rate==1
  
    // eoGenerationalReplacement<Indi> replace; 
    // eoWeakElitistReplacement<Indi> replace(replace_main); 
  
    double elita_percentage = 0.05; 
    eoDeterministicSaDReplacement<Indi> replace(elita_percentage,1.0-elita_percentage);


 
    //////////////////////////////////////
    // The variation operators
    //////////////////////////////////////
    eo1PtBitXover<Indi> xover;
    eoBitMutation<Indi>  mutation(P_MUT_PER_BIT);

    eoDynSGATransform<Indi> transform(xover, &CROSS_RATE, mutation, &MUT_RATE);


    eoSelectTransform<Indi> selectTransform(select, transform);

    //////////////////////////////////////
    // termination condition
    /////////////////////////////////////
    

    eoCheckPoint<Indi> checkpoint(gui);    

     eoValueParam<unsigned> generationCounter(0, "Gen.");
     eoIncrementor<unsigned> increment(generationCounter.value());
     checkpoint.add(increment);

     eoBestFitnessStat<Indi> bestStat;
     eoSecondMomentStats<Indi> SecondStat;
     checkpoint.add(bestStat);
     checkpoint.add(SecondStat);
    
     eoFileMonitor fileMonitor(saveStatsName, " ");
     checkpoint.add(fileMonitor);
     fileMonitor.add(generationCounter);
     fileMonitor.add(bestStat);
     fileMonitor.add(SecondStat);

   
     
     /////////////////////////////////////////
     // the algorithm
     ////////////////////////////////////////
     eoEasyEA<Indi> gga(checkpoint, 
			eval, 
			selectTransform, 
			replace);


     // Apply algo to pop - that's it!
     gga(pop);
  
     // OUTPUT
     pop.sort();
     cout << "FINAL Population\n" << pop << endl;

     IndiStrategy best( pop[0] );
     best.print(); 

}
 

int main(int argc, char **argv)
{

    try
    {
        main_function(argc, argv);
    }
    catch(exception& e)
    {
        cout << "Exception: " << e.what() << '\n';
    }

    return 1;
}
